/*
 * Rubik's cube interface
 *
 * Copyright (C) 2007, 2008 Emmanuel Thome'
 * <Emmanuel dot Thome at normalesup dot org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * bare skeleton: Copyright (C) 1999 Chris Abernethy
 * <chris65536@home.com> (GPL2+)
 */
/* code also taken from gtkglarea/trunk/examples/viewlw.c:
 * Copyright (C) 1998 Janne Lo"f <jlof@mail.student.oulu.fi>
 * GPL
 */

#include <GL/gl.h>
#include <GL/glu.h>
#include <glib.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include <gtkgl/gtkglarea.h>
#include <iostream>
#include <cstring>
#include <cstdlib>
#include <cassert>

#include "manu.h"
#include "trackball.h"
#include "cube.hpp"
#include "rubik.hpp"



GtkWidget *create_glarea(void);
gint glarea_button_release(GtkWidget *, GdkEventButton *);
gint glarea_button_press(GtkWidget *, GdkEventButton *);
gint glarea_scroll(GtkWidget *, GdkEventScroll *);
gint glarea_motion_notify(GtkWidget *, GdkEventMotion *);
gint glarea_draw(GtkWidget *, GdkEventExpose *);
gint glarea_reshape(GtkWidget *, GdkEventConfigure *);
gint glarea_init(GtkWidget *);
gint glarea_destroy(GtkWidget *);
int main(int, char **);

GLfloat cube_angle=0.0;
GLfloat current_move_rot=0.0;

float begin_std_x;
float begin_std_y;
float quat[4];
float zoom;
GLfloat rot[4][4];

#define	MOVING_X1	0x01
#define	MOVING_X2	0x02
#define	MOVING_Y1	0x04
#define	MOVING_Y2	0x08
#define	MOVING_Z1	0x10
#define	MOVING_Z2	0x20

#define	MOVING_CLOCKWISE	0x80

#define	MOVING_T	MOVING_Y2
#define	MOVING_C	MOVING_Z2
#define	MOVING_R	MOVING_X2
#define	MOVING_F	MOVING_Z1
#define	MOVING_L	MOVING_X1
#define	MOVING_B	MOVING_Y1

#define	MOVING_X	(MOVING_X1 | MOVING_X2)
#define	MOVING_Y	(MOVING_Y1 | MOVING_Y2)
#define	MOVING_Z	(MOVING_Z1 | MOVING_Z2)

#define	MOVING_MASK	(MOVING_X|MOVING_Y|MOVING_Z|MOVING_CLOCKWISE)

const char *face_names[12] = {
	"T", "C", "R", "F", "L", "B",
	"-T", "-C", "-R", "-F", "-L", "-B",
};
int button_order[6] = { 0, 5, 1, 3, 4, 2 };

const unsigned int move_ids[6]={
	MOVING_T, MOVING_C, MOVING_R,
	MOVING_F, MOVING_L, MOVING_B,
};

class move {
public:
	unsigned int id;
	int nstep;
	move(int m) :
		id(move_ids[ABS(m)-1] ^ ((m<0)?MOVING_CLOCKWISE:0)), nstep(0) {}
};

#define MOVE_STEPS	20

#include <vector>
#include <queue>

std::deque<move> pending_moves;

std::deque<move> solution;

#define	MAX_MOVE_QUEUE	4
/*
unsigned int npending_moves=0;
unsigned int current_move=0;
unsigned int next_move=0;

struct move pending_moves[MAX_MOVE_QUEUE];

*/
unsigned int moving=0;

typedef double rgbcol[3];

rgbcol couleurs[6] = {
	{ 1,1,1 },
	{ 0.9,0,0 },
	{ 1,1,0 },
	{ 1,0.5,0 },
	{ 0,1,0 },
	{ 0,0,1 },
};

int facettes[54] =
	{ 	0,0,0,0,0,0,0,0,0,	/* dessus */
		1,1,1,1,1,1,1,1,1,	/* avant */
		2,2,2,2,2,2,2,2,2,	/* droite */
		3,3,3,3,3,3,3,3,3,	/* arriere */
		4,4,4,4,4,4,4,4,4,	/* gauche */
		5,5,5,5,5,5,5,5,5,	/* bas */
	};

void schedule_move(unsigned int m)
{
	/*
	if (pending_moves.size() == MAX_MOVE_QUEUE) {
		g_print("Move not taken ; too many in queue\n");
		return;
	}
	*/
	pending_moves.push_back(move(m));
	solution.push_front(move(-m));
	
	/*
	unsigned int pos;

	ASSERT((moving==0) == (current_move == next_move));

	if (npending_moves==(MAX_MOVE_QUEUE-(moving!=0))) {
		g_print("Move not taken ; too many in queue\n");
		return;
	}
	
	pos = (next_move + npending_moves) % MAX_MOVE_QUEUE;

	pending_moves[pos].id=move;
	pending_moves[pos].nstep=0;
	npending_moves++;
	*/
}

void advance_on_solution(GtkWidget * widget)
{
	if (!solution.empty()) {
		move m=solution.front();
		pending_moves.push_back(m);
		solution.pop_front();
	}
}

void do_all_moves(GtkWidget * widget)
{
	for(;!solution.empty();) {
		advance_on_solution(widget);
	}
}

float map_to_standard(double x, double w)
{
	return (2.0 * x - w) /  (float) w;
}

float crossprod(float xa, float ya, float xb, float yb, float xc, float yc)
{
	return (xb-xa) * (yc-ya) - (xc-xa) * (yb-ya);
}

#define triangle_is_direct(xa,ya,xb,yb,xc,yc)	(crossprod(xa,ya,xb,yb,xc,yc)>0)

struct face_status {
	int visible;
	float xa, ya;
	float xb, yb;
	float xc, yc;
	float xd, yd;
	float coord_mat[4];
};

struct face_status visi[6];

void update_visi()
{
	int i;
	for(i=0;i<6;i++) {
		float d = crossprod(
				visi[i].xa, visi[i].ya,
				visi[i].xb, visi[i].yb,
				visi[i].xd, visi[i].yd);
		if ((visi[i].visible = d>0.0)) {
			/* compute the base change matrix */
			visi[i].coord_mat[0]=(visi[i].yd-visi[i].ya)/d;
			visi[i].coord_mat[1]=-(visi[i].yb-visi[i].ya)/d;
			visi[i].coord_mat[2]=-(visi[i].xd-visi[i].xa)/d;
			visi[i].coord_mat[3]=(visi[i].xb-visi[i].xa)/d;
		}
		/*
		g_print("Face %d ; A=(%.0f,%.0f) B=(%.0f,%.0f) C=(%.0f,%.0f) D=(%.0f,%.0f) %svisible\n",
				i,
				visi[i].xa, visi[i].ya,
				visi[i].xb, visi[i].yb,
				visi[i].xc, visi[i].yc,
				visi[i].xd, visi[i].yd,
				visi[i].visible?"":"in");
				*/
	}
}

int in_face(float x, float y, int * tile)
{
	int res=-1;
	int i;
	for(i=0;i<6;i++) {
		float p,q;

		if (!visi[i].visible) continue;
		if (!triangle_is_direct(x,y,
					visi[i].xa, visi[i].ya,
					visi[i].xb, visi[i].yb)) continue;
		if (!triangle_is_direct(x,y,
					visi[i].xb, visi[i].yb,
					visi[i].xc, visi[i].yc)) continue;
		if (!triangle_is_direct(x,y,
					visi[i].xc, visi[i].yc,
					visi[i].xd, visi[i].yd)) continue;
		if (!triangle_is_direct(x,y,
					visi[i].xd, visi[i].yd,
					visi[i].xa, visi[i].ya)) continue;
		if (res!=-1) {
			g_print("Error : both on faces %d and %d !\n",res,i);
		}
		res=i;

		p=(x-visi[i].xa) * visi[i].coord_mat[0] + (y-visi[i].ya) * visi[i].coord_mat[2];
		q=(x-visi[i].xa) * visi[i].coord_mat[1] + (y-visi[i].ya) * visi[i].coord_mat[3];

		*tile = 3*(int)(3*p) + (int)(3*q);
		if (*tile < 0) *tile=0;
		if (*tile > 8) *tile=8;
		/*
		g_print("Coordinates on face: (%.2f, %.2f) -> tile %d\n",
				p,q, *tile);
				*/
	}
	return res;
}

gint key_press_event(GtkWidget * widget, GdkEventKey * event)
{
	/* g_print("Got keypress\n"); */
	switch (event->keyval) {
	case GDK_q:
		gtk_main_quit();
		break;
	}
	// gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "key_press_event");
	return TRUE;
}

gint key_release_event(GtkWidget * widget, GdkEventKey * event)
{
	/* g_print("Got keyrelease\n"); */
	// gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "key_release_event");
	return TRUE;
}

GtkWidget *create_glarea(void)
{

	GtkWidget *glarea;

	int attrlist[] = {
		GDK_GL_RGBA,
		GDK_GL_DOUBLEBUFFER,
		GDK_GL_DEPTH_SIZE, 1,
		GDK_GL_NONE
	};

	if (gdk_gl_query() == FALSE) {
		g_print("OpenGL not supported!\n");
		return NULL;
	}

	if ((glarea = gtk_gl_area_new(attrlist)) == NULL) {
		g_print("Error creating GtkGLArea!\n");
		return NULL;
	}

	gtk_widget_set_events(GTK_WIDGET(glarea),
			      GDK_EXPOSURE_MASK |
			      GDK_KEY_PRESS_MASK |
			      GDK_KEY_RELEASE_MASK |
			      GDK_BUTTON_PRESS_MASK |
			      GDK_BUTTON_RELEASE_MASK |
			      GDK_POINTER_MOTION_MASK |
			      GDK_POINTER_MOTION_HINT_MASK |
                              GDK_BUTTON4_MASK |
                              GDK_BUTTON5_MASK);

	gtk_signal_connect(GTK_OBJECT(glarea), "button_release_event",
			   GTK_SIGNAL_FUNC(glarea_button_release), NULL);

	gtk_signal_connect(GTK_OBJECT(glarea), "button_press_event",
			   GTK_SIGNAL_FUNC(glarea_button_press), NULL);

	gtk_signal_connect(GTK_OBJECT(glarea), "scroll_event",
			   GTK_SIGNAL_FUNC(glarea_scroll), NULL);

	gtk_signal_connect(GTK_OBJECT(glarea), "motion_notify_event",
			   GTK_SIGNAL_FUNC(glarea_motion_notify), NULL);

	gtk_signal_connect(GTK_OBJECT(glarea), "expose_event",
			   GTK_SIGNAL_FUNC(glarea_draw), NULL);

	gtk_signal_connect(GTK_OBJECT(glarea), "configure_event",
			   GTK_SIGNAL_FUNC(glarea_reshape), NULL);

	gtk_signal_connect(GTK_OBJECT(glarea), "realize",
			   GTK_SIGNAL_FUNC(glarea_init), NULL);

	gtk_signal_connect(GTK_OBJECT(glarea), "destroy",
			   GTK_SIGNAL_FUNC(glarea_destroy), NULL);

	gtk_widget_set_usize(GTK_WIDGET(glarea), 512,512);

	begin_std_x=0;
	begin_std_y=0;
	zoom=45;
	trackball(quat,0,0,0,0);



	return (glarea);

}

gint glarea_button_release(GtkWidget * widget, GdkEventButton * event)
{
	/* NB ; on ment sur le système de coordonnées ! */
	if (event->button) {
		// int h = widget->allocation.height;
		// int x = event->x;
		// int y = event->y;
		// g_print("Button %d release (%d, %d)\n", event->button, x, h-y);
		return TRUE;
	}

	return FALSE;

}

gint handle_button(GtkWidget * widget, gpointer data)
{
	int t;

	t = (char **) data - (char **) face_names;
	/*
	int s;
	s = (t>=6)?MOVING_CLOCKWISE:0;
	t = t % 6;
	*/

	/* g_print("Button %s clicked\n", face_names[t]); */
	schedule_move(((t>=6)?-1:1)*(1+(t%6)));
			/*
	switch(t) {
		case 0:	schedule_move(s | MOVING_T); break;
		case 1:	schedule_move(s | MOVING_C); break;
		case 2:	schedule_move(s | MOVING_R); break;
		case 3:	schedule_move(s | MOVING_F); break;
		case 4:	schedule_move(s | MOVING_L); break;
		case 5:	schedule_move(s | MOVING_B); break;
	}
	*/

	return TRUE;
}

gint glarea_scroll(GtkWidget * widget, GdkEventScroll * event)
{
        GdkModifierType state = (GdkModifierType) event->state;

	if (state & GDK_BUTTON4_MASK || event->direction == GDK_SCROLL_UP) {
		zoom += 3;
		gtk_widget_draw(widget, NULL);
                return TRUE;
	}

	if (state & GDK_BUTTON5_MASK || event->direction == GDK_SCROLL_DOWN) {
		zoom -= 3;
		gtk_widget_draw(widget, NULL);
                return TRUE;
	}

        return FALSE;
}

gint glarea_button_press(GtkWidget * widget, GdkEventButton * event)
{
        if (!(event->button))
                return FALSE;

        /* don't respond to double clicks. */
        if (event->type != GDK_BUTTON_PRESS) {
                return FALSE;
        }

        int w = widget->allocation.width;
        int h = widget->allocation.height;
        double x = event->x;
        double y = event->y;
        int f;
        int t = 0;

        /* NB ; on ment sur le système de coordonnées ! */
        /*
           g_print("Button %d press (t=%d)  (%d, %d)\n",
           event->button, event->type,
           x, h-y);
           */

        f=in_face(x,h-y,&t);

        if (f!=-1) {

                if (t==4) {
                        g_print("This is face %s\n", face_names[f]);
                } else if (event->button == 3) {
                        t+=f*9;
                        facettes[t]=(facettes[t]+1)%6;
                        solution.clear();
                        gtk_widget_draw(GTK_WIDGET(widget), NULL);
                }
        }

        /* memorize this, as the drag starts */
        begin_std_x=map_to_standard(x,w);
        begin_std_y=-map_to_standard(y,h);
        return TRUE;
}

/* code taken from gtkglarea/trunk/examples/viewlw.c:
 * Copyright (C) 1998 Janne Löf <jlof@mail.student.oulu.fi>
 * GPL
 */
gint glarea_motion_notify(GtkWidget * widget, GdkEventMotion * event)
{
	float nx,ny;
	double x;
	double y;
	int w = widget->allocation.width;
	int h = widget->allocation.height;

	GdkModifierType state;

	if (event->is_hint) {
		int zx,zy;
		gdk_window_get_pointer(event->window, &zx, &zy, &state);
		x=zx;
		y=zy;
	} else {
		x = event->x;
		y = event->y;
		state = (GdkModifierType) event->state;
	}

	nx=map_to_standard(x,w);
	ny=-map_to_standard(y,h);

	/* NB ; on ment sur le système de coordonnées ! */
	if (state & GDK_BUTTON1_MASK) {
		// g_print("Button 1 motion  (%d, %d)\n", x, h-y);
		float spin[4];
		trackball(spin,begin_std_x,begin_std_y,nx,ny);
		add_quats(spin,quat,quat);
		gtk_widget_draw(widget, NULL);
	}

	if (state & GDK_BUTTON2_MASK) {
		// g_print("Button 2 motion  (%d, %d)\n", x, h-y);
		zoom += (ny-begin_std_y)*40;
		if (zoom<5) zoom=5;
		if (zoom>120) zoom=120;
		gtk_widget_draw(widget, NULL);
		// g_print("Zoom is now %.2f\n", zoom);
	}

	if (state & GDK_BUTTON3_MASK) {
		// g_print("Button 3 motion  (%d, %d)\n", x, h-y);
	}

	begin_std_x=nx;
	begin_std_y=ny;


	return TRUE;

}

void couleur_facette(int x)
{
	rgbcol c;
	memcpy(c,couleurs[facettes[x]],sizeof(rgbcol));
	glColor3d(c[0],c[1],c[2]);
}

typedef	int tile_change[54];

tile_change face_effect[6] = {
	{ /* T */ 2, 5, 8, 1, 4, 7, 0, 3, 6, 36, 37, 38, 12, 13, 14, 15, 16,
	 17, 9, 10, 11, 21, 22, 23, 24, 25, 26, 18, 19, 20, 30, 31, 32, 33,
	 34, 35, 27, 28, 29, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50,
	 51, 52, 53} ,
	{ /* C */ 0, 1, 2, 3, 4, 5, 18, 21, 24, 11, 14, 17, 10, 13, 16, 9, 12,
	 15, 47, 19, 20, 46, 22, 23, 45, 25, 26, 27, 28, 29, 30, 31, 32, 33,
	 34, 35, 36, 37, 8, 39, 40, 7, 42, 43, 6, 38, 41, 44, 48, 49, 50, 51,
	 52, 53} ,
	{ /* R */ 0, 1, 33, 3, 4, 30, 6, 7, 27, 9, 10, 2, 12, 13, 5, 15, 16,
	 8, 20, 23, 26, 19, 22, 25, 18, 21, 24, 53, 28, 29, 50, 31, 32, 47,
	 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 11, 48, 49, 14,
	 51, 52, 17} ,
	{ /* F */ 42, 39, 36, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16,
	 17, 18, 19, 0, 21, 22, 1, 24, 25, 2, 29, 32, 35, 28, 31, 34, 27, 30,
	 33, 51, 37, 38, 52, 40, 41, 53, 43, 44, 45, 46, 47, 48, 49, 50, 26,
	 23, 20} ,
	{ /* L */ 9, 1, 2, 12, 4, 5, 15, 7, 8, 45, 10, 11, 48, 13, 14, 51, 16,
	 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 6, 30, 31, 3, 33,
	 34, 0, 38, 41, 44, 37, 40, 43, 36, 39, 42, 35, 46, 47, 32, 49, 50,
	 29, 52, 53} ,
	{ /* B */ 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 24, 25,
	 26, 18, 19, 20, 21, 22, 23, 33, 34, 35, 27, 28, 29, 30, 31, 32, 42,
	 43, 44, 36, 37, 38, 39, 40, 41, 15, 16, 17, 47, 50, 53, 46, 49, 52,
	 45, 48, 51} ,
};

void apply_perm(unsigned int id)
{
	int i;
	int anciennes_facettes[54];
	int k=-1;

	ASSERT(id!=0);

	if (id & MOVING_Y2)	k=0;
	if (id & MOVING_Y1)	k=5;
	if (id & MOVING_Z2)	k=1;
	if (id & MOVING_Z1)	k=3;
	if (id & MOVING_X1)	k=4;
	if (id & MOVING_X2)	k=2;

	ASSERT(k!=-1);

	memcpy(anciennes_facettes,facettes,54*sizeof(int));
	if (id & MOVING_CLOCKWISE) {
		for(i=0;i<54;i++) {
			facettes[i]=anciennes_facettes[face_effect[k][i]];
		}
	} else {
		for(i=0;i<54;i++) {
			facettes[face_effect[k][i]]=anciennes_facettes[i];
		}
	}
}

typedef int e_hint_t[4];

const e_hint_t edge_hints[12] = {
	{ 0,3,    1,28 },
	{ 0,2,    5,19 },
	{ 0,1,    7,10 },
	{ 0,4,    3,37 },

	{ 4,1,   41,12 },
	{ 1,2,   14,21 },
	{ 2,3,   23,30 },
	{ 3,4,   32,39 },

	{ 1,5,   16,46 },
	{ 2,5,   25,50 },
	{ 3,5,   34,52 },
	{ 4,5,   43,48 },
};

int recognize_edge(int * pa, int * pb, int ca, int cb)
{
	const e_hint_t * h = edge_hints;
	int i;
	for(i=0;i<12;i++) {
		if (ca==h[i][0] && cb==h[i][1]) {
			*pa=h[i][2];
			*pb=h[i][3];
			return 1;
		} else if (ca==h[i][1] && cb==h[i][0]) {
			*pa=h[i][3];
			*pb=h[i][2];
			return 1;
		}
	}
	return 0;
}


/* We always turn counter-clockwise around the vertices */
typedef	int v_hint_t[6];
const v_hint_t vertex_hints[8] = {
	{ 0,4,1,    6,38, 9 },
	{ 0,1,2,    8,11,18 },
	{ 0,2,3,    2,20,27 },
	{ 0,3,4,    0,29,36 },
	{ 5,1,4,   45,15,44 },
	{ 5,4,3,   51,42,35 },
	{ 5,3,2,   53,33,26 },
	{ 5,2,1,   47,24,17 },
};
int recognize_vertex(int * pa, int * pb, int * pc, int ca, int cb, int cc)
{
	const v_hint_t * h = vertex_hints;
	int i;
	for(i=0;i<8;i++) {
		if (ca==h[i][0] && cb==h[i][1] && cc == h[i][2]) {
			*pa=h[i][3];
			*pb=h[i][4];
			*pc=h[i][5];
			return 1;
		} else if (ca==h[i][1] && cb==h[i][2] && cc == h[i][0]) {
			*pa=h[i][4];
			*pb=h[i][5];
			*pc=h[i][3];
			return 1;
		} else if (ca==h[i][2] && cb==h[i][0] && cc == h[i][1]) {
			*pa=h[i][5];
			*pb=h[i][3];
			*pc=h[i][4];
			return 1;
		}
	}
	return 0;
}

int recognize_perm(tile_change dest)
{
	int i;
	int b;

	for(i=0;i<54;dest[i++]=-1);

	/* register all the fixed points */
	for(i=4;i<54;i+=9) dest[i]=i;

	for(i=0;i<12;i++) {
		int es[2];
		int e[2];
		es[0]=edge_hints[i][2];
		es[1]=edge_hints[i][3];
		b=recognize_edge(&e[0],&e[1], facettes[es[0]], facettes[es[1]]);
		if (!b) {
			g_print("Error: edge at %s/%s is illegal\n",
					face_names[es[0]/9],
					face_names[es[1]/9]);
			return 0;
		} else if (dest[e[0]]!=-1 || dest[e[1]]!=-1) {
			g_print("Error: edge %s/%s found twice (last: %s/%s)\n",
					face_names[e[0]/9], 
					face_names[e[1]/9],
					face_names[es[0]/9], 
					face_names[es[1]/9]);
			return 0;
		}
		/*
		g_print("Found edge %s/%s at %s/%s\n",
				face_names[e[0]/9],
				face_names[e[1]/9],
				face_names[es[0]/9],
				face_names[es[1]/9]);
				*/
		dest[e[0]]=es[0];
		dest[e[1]]=es[1];
	}
	for(i=0;i<8;i++) {
		int vs[3];
		int v[3];
		vs[0]=vertex_hints[i][3];
		vs[1]=vertex_hints[i][4];
		vs[2]=vertex_hints[i][5];
		b=recognize_vertex(&v[0],&v[1],&v[2],
				facettes[vs[0]],
				facettes[vs[1]],
				facettes[vs[2]]);
		if (!b) {
			g_print("Error: vertex at %s/%s/%s is illegal\n",
					face_names[vs[0]/9],
					face_names[vs[1]/9],
					face_names[vs[2]/9]);
			return 0;
		} else if (dest[v[0]]!=-1 || dest[v[1]]!=-1 || dest[v[2]]!=-1) {
			g_print("Error: vertex %s/%s/%s found twice "
					"(last: %s/%s/%s)\n",
					face_names[v[0]/9], 
					face_names[v[1]/9],
					face_names[v[2]/9],
					face_names[vs[0]/9], 
					face_names[vs[1]/9], 
					face_names[vs[2]/9]);
			return 0;
		}
		/*
		g_print("Found vertex %s/%s/%s at %s/%s/%s\n",
				face_names[v[0]/9], 
				face_names[v[1]/9],
				face_names[v[2]/9],
				face_names[vs[0]/9], 
				face_names[vs[1]/9], 
				face_names[vs[2]/9]);
				*/
		dest[v[0]]=vs[0];
		dest[v[1]]=vs[1];
		dest[v[2]]=vs[2];
	}
#if 0
	g_print("Succesfully identified permutation of tiles\n");
	/* which does not mean that it is an admissible element of the
	 * rubik group !*/
	g_print("[");
	for(i=0;i<54;i++) g_print(" %d", dest[i]);
	g_print(" ]\n");
#endif
	return 1;
}

gmp_randstate_t random_source;

void randomize(GtkWidget * widget, GtkWidget * glarea)
{
	if (moving || pending_moves.size())
		return;

	perm<48> p = rubik_sift.random(random_source);
	std::vector<int> xvec = corresp_48_to_54(p);
	solution.clear();

	int anciennes_facettes[54];
	memcpy(anciennes_facettes,facettes,54*sizeof(int));
	for(int i=0;i<54;i++) {
		facettes[i]=anciennes_facettes[xvec[i]];
	}

	gtk_widget_draw(GTK_WIDGET(glarea), NULL);
}
	
void solve_cube(GtkWidget * widget)
{
	tile_change x;

	if (!recognize_perm(x)) {
		std::cout << "This configuration is illegal\n";
		return;
	}

	std::vector<int> xvec;

	xvec.insert(xvec.begin(),x,x+54);

	perm<48> p = corresp_54_to_48(xvec);

	// std::cout << "This is == " << p << "\n";

	std::vector<int> words;
	bool b;

	// rubik_sift.loud();

	b=rubik_sift.inverse(p,words);
	if (!b) {
		std::cout << "No inverse found ; element not in group\n";
	} else {
		std::cout << "Solution found: " << words.size() <<  " moves\n";
		// std::cout << "Solution found: " << words <<  "\n";
		solution.assign(words.begin(),words.end());
	}
}

GLuint tex[6];

unsigned char * texture;

#define	TEX_SIZE	64
#define	TEX_BORDER	4

void mktex(rgbcol c)
{
	int i,j,k=0;
	unsigned char s[3];
	texture=(unsigned char *)malloc(TEX_SIZE * TEX_SIZE * 3);
	s[0]=(unsigned char) 255 * (int) c[0];
	s[1]=(unsigned char) 255 * (int) c[1];
	s[2]=(unsigned char) 255 * (int) c[2];
	k=0;
	memset(texture,0,TEX_SIZE * TEX_SIZE * 3);

	for(i=TEX_BORDER;i<TEX_SIZE-TEX_BORDER;i++) {
		k=(i*TEX_SIZE + TEX_BORDER) * 3;
		for(j=TEX_BORDER;j<TEX_SIZE-TEX_BORDER;j++) {
			texture[k++]=s[0];
			texture[k++]=s[1];
			texture[k++]=s[2];
		}
	}
}

/* Here come many macros ; I could have done this more conceptually, for
 * sure.
 */

// 16 control points on each face. Numbered top to bottom, left to right,
// from 0 to 15.
#define	TH1	0.333333333
// top
#define	T0	-1.0, 1.0,-1.0
#define	T1	-TH1, 1.0,-1.0
#define	T2	 TH1, 1.0,-1.0
#define	T3	 1.0, 1.0,-1.0
#define	T4	-1.0, 1.0,-TH1
#define	T5	-TH1, 1.0,-TH1
#define	T6	 TH1, 1.0,-TH1
#define	T7	 1.0, 1.0,-TH1
#define	T8	-1.0, 1.0, TH1
#define	T9	-TH1, 1.0, TH1
#define	T10	 TH1, 1.0, TH1
#define	T11	 1.0, 1.0, TH1
#define	T12	-1.0, 1.0, 1.0
#define	T13	-TH1, 1.0, 1.0
#define	T14	 TH1, 1.0, 1.0
#define	T15	 1.0, 1.0, 1.0
// close (front)
#define	C0	-1.0, 1.0, 1.0
#define	C1	-TH1, 1.0, 1.0
#define	C2	 TH1, 1.0, 1.0
#define	C3	 1.0, 1.0, 1.0
#define	C4	-1.0, TH1, 1.0
#define	C5	-TH1, TH1, 1.0
#define	C6	 TH1, TH1, 1.0
#define	C7	 1.0, TH1, 1.0
#define	C8	-1.0,-TH1, 1.0
#define	C9	-TH1,-TH1, 1.0
#define	C10	 TH1,-TH1, 1.0
#define	C11	 1.0,-TH1, 1.0
#define	C12	-1.0,-1.0, 1.0
#define	C13	-TH1,-1.0, 1.0
#define	C14	 TH1,-1.0, 1.0
#define	C15	 1.0,-1.0, 1.0
// right
#define	R0	 1.0, 1.0, 1.0
#define	R1	 1.0, 1.0, TH1
#define	R2	 1.0, 1.0,-TH1
#define	R3	 1.0, 1.0,-1.0
#define	R4	 1.0, TH1, 1.0
#define	R5	 1.0, TH1, TH1
#define	R6	 1.0, TH1,-TH1
#define	R7	 1.0, TH1,-1.0
#define	R8	 1.0,-TH1, 1.0
#define	R9	 1.0,-TH1, TH1
#define	R10	 1.0,-TH1,-TH1
#define	R11	 1.0,-TH1,-1.0
#define	R12	 1.0,-1.0, 1.0
#define	R13	 1.0,-1.0, TH1
#define	R14	 1.0,-1.0,-TH1
#define	R15	 1.0,-1.0,-1.0
// far (back)
#define	F0	 1.0, 1.0,-1.0
#define	F1	 TH1, 1.0,-1.0
#define	F2	-TH1, 1.0,-1.0
#define	F3	-1.0, 1.0,-1.0
#define	F4	 1.0, TH1,-1.0
#define	F5	 TH1, TH1,-1.0
#define	F6	-TH1, TH1,-1.0
#define	F7	-1.0, TH1,-1.0
#define	F8	 1.0,-TH1,-1.0
#define	F9	 TH1,-TH1,-1.0
#define	F10	-TH1,-TH1,-1.0
#define	F11	-1.0,-TH1,-1.0
#define	F12	 1.0,-1.0,-1.0
#define	F13	 TH1,-1.0,-1.0
#define	F14	-TH1,-1.0,-1.0
#define	F15	-1.0,-1.0,-1.0
// left
#define	L0	-1.0, 1.0,-1.0
#define	L1	-1.0, 1.0,-TH1
#define	L2	-1.0, 1.0, TH1
#define	L3	-1.0, 1.0, 1.0
#define	L4	-1.0, TH1,-1.0
#define	L5	-1.0, TH1,-TH1
#define	L6	-1.0, TH1, TH1
#define	L7	-1.0, TH1, 1.0
#define	L8	-1.0,-TH1,-1.0
#define	L9	-1.0,-TH1,-TH1
#define	L10	-1.0,-TH1, TH1
#define	L11	-1.0,-TH1, 1.0
#define	L12	-1.0,-1.0,-1.0
#define	L13	-1.0,-1.0,-TH1
#define	L14	-1.0,-1.0, TH1
#define	L15	-1.0,-1.0, 1.0
// bottom
#define	B0	-1.0,-1.0, 1.0
#define	B1	-TH1,-1.0, 1.0
#define	B2	 TH1,-1.0, 1.0
#define	B3	 1.0,-1.0, 1.0
#define	B4	-1.0,-1.0, TH1
#define	B5	-TH1,-1.0, TH1
#define	B6	 TH1,-1.0, TH1
#define	B7	 1.0,-1.0, TH1
#define	B8	-1.0,-1.0,-TH1
#define	B9	-TH1,-1.0,-TH1
#define	B10	 TH1,-1.0,-TH1
#define	B11	 1.0,-1.0,-TH1
#define	B12	-1.0,-1.0,-1.0
#define	B13	-TH1,-1.0,-1.0
#define	B14	 TH1,-1.0,-1.0
#define	B15	 1.0,-1.0,-1.0

#define	MAP(i,a,b,c,d)	\
		couleur_facette(i);		\
		glTexCoord2f(0,1); glVertex3f(a);	\
		glTexCoord2f(0,0); glVertex3f(b);	\
		glTexCoord2f(1,0); glVertex3f(c);	\
		glTexCoord2f(1,1); glVertex3f(d);
// all tiles.
#define	TILE0()		MAP( 0, T1, T0, T4, T5)
#define	TILE1()		MAP( 1, T2, T1, T5, T6)
#define	TILE2()		MAP( 2, T3, T2, T6, T7)
#define	TILE3()		MAP( 3, T5, T4, T8, T9)
#define	TILE4()		MAP( 4, T6, T5, T9,T10)
#define	TILE5()		MAP( 5, T7, T6,T10,T11)
#define	TILE6()		MAP( 6, T9, T8,T12,T13)
#define	TILE7()		MAP( 7,T10, T9,T13,T14)
#define	TILE8()		MAP( 8,T11,T10,T14,T15)
#define	TILE9()		MAP( 9, C1, C0, C4, C5)
#define	TILE10()	MAP(10, C2, C1, C5, C6)
#define	TILE11()	MAP(11, C3, C2, C6, C7)
#define	TILE12()	MAP(12, C5, C4, C8, C9)
#define	TILE13()	MAP(13, C6, C5, C9,C10)
#define	TILE14()	MAP(14, C7, C6,C10,C11)
#define	TILE15()	MAP(15, C9, C8,C12,C13)
#define	TILE16()	MAP(16,C10, C9,C13,C14)
#define	TILE17()	MAP(17,C11,C10,C14,C15)
#define	TILE18()	MAP(18, R1, R0, R4, R5)
#define	TILE19()	MAP(19, R2, R1, R5, R6)
#define	TILE20()	MAP(20, R3, R2, R6, R7)
#define	TILE21()	MAP(21, R5, R4, R8, R9)
#define	TILE22()	MAP(22, R6, R5, R9,R10)
#define	TILE23()	MAP(23, R7, R6,R10,R11)
#define	TILE24()	MAP(24, R9, R8,R12,R13)
#define	TILE25()	MAP(25,R10, R9,R13,R14)
#define	TILE26()	MAP(26,R11,R10,R14,R15)
#define	TILE27()	MAP(27, F1, F0, F4, F5)
#define	TILE28()	MAP(28, F2, F1, F5, F6)
#define	TILE29()	MAP(29, F3, F2, F6, F7)
#define	TILE30()	MAP(30, F5, F4, F8, F9)
#define	TILE31()	MAP(31, F6, F5, F9,F10)
#define	TILE32()	MAP(32, F7, F6,F10,F11)
#define	TILE33()	MAP(33, F9, F8,F12,F13)
#define	TILE34()	MAP(34,F10, F9,F13,F14)
#define	TILE35()	MAP(35,F11,F10,F14,F15)
#define	TILE36()	MAP(36, L1, L0, L4, L5)
#define	TILE37()	MAP(37, L2, L1, L5, L6)
#define	TILE38()	MAP(38, L3, L2, L6, L7)
#define	TILE39()	MAP(39, L5, L4, L8, L9)
#define	TILE40()	MAP(40, L6, L5, L9,L10)
#define	TILE41()	MAP(41, L7, L6,L10,L11)
#define	TILE42()	MAP(42, L9, L8,L12,L13)
#define	TILE43()	MAP(43,L10, L9,L13,L14)
#define	TILE44()	MAP(44,L11,L10,L14,L15)
#define	TILE45()	MAP(45, B1, B0, B4, B5)
#define	TILE46()	MAP(46, B2, B1, B5, B6)
#define	TILE47()	MAP(47, B3, B2, B6, B7)
#define	TILE48()	MAP(48, B5, B4, B8, B9)
#define	TILE49()	MAP(49, B6, B5, B9,B10)
#define	TILE50()	MAP(50, B7, B6,B10,B11)
#define	TILE51()	MAP(51, B9, B8,B12,B13)
#define	TILE52()	MAP(52,B10, B9,B13,B14)
#define	TILE53()	MAP(53,B11,B10,B14,B15)

/* Now, we define three reading orders ; these are useful for the
 * movement display, as the rotation must be put in the right place.
 *
 * each reading order has 5 steps.
 */
/* Go from Y=-1 to Y=1 */
#define	READ_Y0()	\
	TILE45(); TILE46(); TILE47();	\
	TILE48(); TILE49(); TILE50();	\
	TILE51(); TILE52(); TILE53()
#define	READ_Y1()	\
	TILE15(); TILE16(); TILE17();	\
	TILE24(); TILE25(); TILE26();	\
	TILE33(); TILE34(); TILE35();	\
	TILE42(); TILE43(); TILE44()
#define	READ_Y2()	\
	TILE12(); TILE13(); TILE14();	\
	TILE21(); TILE22(); TILE23();	\
	TILE30(); TILE31(); TILE32();	\
	TILE39(); TILE40(); TILE41()
#define	READ_Y3()	\
	TILE9(); TILE10(); TILE11();	\
	TILE18(); TILE19(); TILE20();	\
	TILE27(); TILE28(); TILE29();	\
	TILE36(); TILE37(); TILE38()
#define	READ_Y4()	\
	TILE0(); TILE1(); TILE2();	\
	TILE3(); TILE4(); TILE5();	\
	TILE6(); TILE7(); TILE8()
	
/* Go from X=-1 to X=1 */
#define	READ_X0()	\
	TILE36(); TILE37(); TILE38();	\
	TILE39(); TILE40(); TILE41();	\
	TILE42(); TILE43(); TILE44()
#define	READ_X1()	\
	TILE6(); TILE3(); TILE0();	\
	TILE29(); TILE32(); TILE35();	\
	TILE51(); TILE48(); TILE45();	\
	TILE15(); TILE12(); TILE9()
#define	READ_X2()	\
	TILE7(); TILE4(); TILE1();	\
	TILE28(); TILE31(); TILE34();	\
	TILE52(); TILE49(); TILE46();	\
	TILE16(); TILE13(); TILE10()
#define	READ_X3()	\
	TILE8(); TILE5(); TILE2();	\
	TILE27(); TILE30(); TILE33();	\
	TILE53(); TILE50(); TILE47();	\
	TILE17(); TILE14(); TILE11()
#define	READ_X4()	\
	TILE18(); TILE19(); TILE20();	\
	TILE21(); TILE22(); TILE23();	\
	TILE24(); TILE25(); TILE26();

/* Go from Z=-1 to Z=1 */
#define	READ_Z0()	\
	TILE27(); TILE28(); TILE29();	\
	TILE30(); TILE31(); TILE32();	\
	TILE33(); TILE34(); TILE35()
#define READ_Z1()	\
	TILE0(); TILE1(); TILE2();	\
	TILE20(); TILE23(); TILE26();	\
	TILE53(); TILE52(); TILE51();	\
	TILE42(); TILE39(); TILE36()
#define READ_Z2()	\
	TILE3(); TILE4(); TILE5();	\
	TILE19(); TILE22(); TILE25();	\
	TILE50(); TILE49(); TILE48();	\
	TILE43(); TILE40(); TILE37()
#define READ_Z3()	\
	TILE6(); TILE7(); TILE8();	\
	TILE18(); TILE21(); TILE24();	\
	TILE47(); TILE46(); TILE45();	\
	TILE44(); TILE41(); TILE38()
#define	READ_Z4()	\
	TILE9(); TILE10(); TILE11();	\
	TILE12(); TILE13(); TILE14();	\
	TILE15(); TILE16(); TILE17()


/* now define cutting panes */
#define	PANE(a,b,c,d)	\
	glColor3f(0,0,0);		\
		glTexCoord2f(0,1); glVertex3f(a);	\
		glTexCoord2f(0,0); glVertex3f(b);	\
		glTexCoord2f(1,0); glVertex3f(c);	\
		glTexCoord2f(1,1); glVertex3f(d);

/* Note that the choice of vertices is somewhat arbitrary */
#define PANE_X1()	PANE(T1,T13,B1,B13)
#define PANE_X2()	PANE(T2,T14,B2,B14)

#define PANE_Y1()	PANE(C8,C11,F8,F11)
#define PANE_Y2()	PANE(C4,C7,F4,F7)

#define	PANE_Z1()	PANE(T4,T7,B11,B8)
#define	PANE_Z2()	PANE(T8,T11,B7,B4)
	
gint glarea_draw(GtkWidget * widget, GdkEventExpose * event)
{

	/* Draw only on the last expose event. */

	if (event->count > 0) {
		return (TRUE);
	}

	/* gtk_gl_area_make_current MUST be called before rendering */
	/* into the GtkGLArea.                                      */

	if (gtk_gl_area_make_current(GTK_GL_AREA(widget))) {
		int s;
		int w = widget->allocation.width;
		int h = widget->allocation.height;

		/* Clear the drawing color buffer and depth buffers */
		/* before drawing.                                  */

		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluPerspective(zoom,(GLfloat)w/(GLfloat)h,0.1f,100.0f);

		glMatrixMode(GL_MODELVIEW);

		glLoadIdentity();
		glTranslatef(0.0f, 0.0f, -8.0f);
		build_rotmatrix(rot,quat);
		glMultMatrixf(&rot[0][0]);
		
		glRotatef(cube_angle,1,0.3,0);

		glBindTexture(GL_TEXTURE_2D,tex[0]);
		/* Note that the matrix stuff _MUST_ be done outside
		 * glBegin()/glEnd()
		 */
		s= (moving & MOVING_CLOCKWISE)?-1:1;
		if (moving & MOVING_X1) {
			// L
			glPushMatrix();
				glRotatef(current_move_rot,s,0,0);
				glBegin(GL_QUADS);
					READ_X0();
					READ_X1();
					PANE_X1();
				glEnd();
			glPopMatrix();
			glBegin(GL_QUADS);
				PANE_X1();
				READ_X2();
				READ_X3();
				READ_X4();
			glEnd();
		} else if (moving & MOVING_X2) {
			// R
			glBegin(GL_QUADS);
				READ_X0();
				READ_X1();
				READ_X2();
				PANE_X2();
			glEnd();
			glPushMatrix();
				glRotatef(current_move_rot,-s,0,0);
				glBegin(GL_QUADS);
					PANE_X2();
					READ_X3();
					READ_X4();
				glEnd();
			glPopMatrix();
		} else if (moving & MOVING_Y1) {
			// T
			glPushMatrix();
				glRotatef(current_move_rot,0,s,0);
				glBegin(GL_QUADS);
					READ_Y0();
					READ_Y1();
					PANE_Y1();
				glEnd();
			glPopMatrix();
			glBegin(GL_QUADS);
				PANE_Y1();
				READ_Y2();
				READ_Y3();
				READ_Y4();
			glEnd();
		} else if (moving & MOVING_Y2) {
			// B
			glBegin(GL_QUADS);
				READ_Y0();
				READ_Y1();
				READ_Y2();
				PANE_Y2();
			glEnd();
			glPushMatrix();
				glRotatef(current_move_rot,0,-s,0);
				glBegin(GL_QUADS);
					PANE_Y2();
					READ_Y3();
					READ_Y4();
				glEnd();
			glPopMatrix();
		} else if (moving & MOVING_Z1) {
			// C
			glPushMatrix();
				glRotatef(current_move_rot,0,0,s);
				glBegin(GL_QUADS);
					READ_Z0();
					READ_Z1();
					PANE_Z1();
				glEnd();
			glPopMatrix();
			glBegin(GL_QUADS);
				PANE_Z1();
				READ_Z2();
				READ_Z3();
				READ_Z4();
			glEnd();
		} else if (moving & MOVING_Z2) {
			// F
			glBegin(GL_QUADS);
				READ_Z0();
				READ_Z1();
				READ_Z2();
				PANE_Z2();
			glEnd();
			glPushMatrix();
				glRotatef(current_move_rot,0,0,-s);
				glBegin(GL_QUADS);
					PANE_Z2();
					READ_Z3();
					READ_Z4();
				glEnd();
			glPopMatrix();
		} else {	/* pick anything */
			glBegin(GL_QUADS);
			READ_Z0();
			READ_Z1();
			READ_Z2();
			READ_Z3();
			READ_Z4();
			glEnd();
		}

		{
			GLdouble model[16];
			GLdouble proj[16];
			GLint view[4];
			GLdouble wx, wy, wz;

			glGetDoublev(GL_PROJECTION_MATRIX,proj);
			glGetDoublev(GL_MODELVIEW_MATRIX,model);
			glGetIntegerv(GL_VIEWPORT,view);

			gluProject( T0,model,proj,view,&wx,&wy,&wz);
			visi[0].xa=wx; visi[0].ya=wy;
			visi[3].xd=wx; visi[3].yd=wy;
			visi[4].xa=wx; visi[4].ya=wy;

			gluProject( T3,model,proj,view,&wx,&wy,&wz);
			visi[0].xd=wx; visi[0].yd=wy;
			visi[2].xd=wx; visi[2].yd=wy;
			visi[3].xa=wx; visi[3].ya=wy;

			gluProject(T12,model,proj,view,&wx,&wy,&wz);
			visi[0].xb=wx; visi[0].yb=wy;
			visi[1].xa=wx; visi[1].ya=wy;
			visi[4].xd=wx; visi[4].yd=wy;

			gluProject(T15,model,proj,view,&wx,&wy,&wz);
			visi[0].xc=wx; visi[0].yc=wy;
			visi[1].xd=wx; visi[1].yd=wy;
			visi[2].xa=wx; visi[2].ya=wy;

			gluProject( B0,model,proj,view,&wx,&wy,&wz);
			visi[5].xa=wx; visi[5].ya=wy;
			visi[1].xb=wx; visi[1].yb=wy;
			visi[4].xc=wx; visi[4].yc=wy;

			gluProject( B3,model,proj,view,&wx,&wy,&wz);
			visi[5].xd=wx; visi[5].yd=wy;
			visi[1].xc=wx; visi[1].yc=wy;
			visi[2].xb=wx; visi[2].yb=wy;

			gluProject(B12,model,proj,view,&wx,&wy,&wz);
			visi[5].xb=wx; visi[5].yb=wy;
			visi[3].xc=wx; visi[3].yc=wy;
			visi[4].xb=wx; visi[4].yb=wy;

			gluProject(B15,model,proj,view,&wx,&wy,&wz);
			visi[5].xc=wx; visi[5].yc=wy;
			visi[2].xc=wx; visi[2].yc=wy;
			visi[3].xb=wx; visi[3].yb=wy;

			update_visi();
		}

		gtk_gl_area_swapbuffers(GTK_GL_AREA(widget));

	}

	return (TRUE);

}

gint glarea_reshape(GtkWidget * widget, GdkEventConfigure * event)
{

	int w = widget->allocation.width;
	int h = widget->allocation.height;

	/* g_print("Reshape Event\n"); */

	if (gtk_gl_area_make_current(GTK_GL_AREA(widget))) {

		glViewport(0, 0, w, h);
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		/*
		// gluOrtho2D(-(w >> 1), (w >> 1), -(h >> 1), h >> 1);
		//
		// bon dieu c'que c'est moche !
		// glOrtho(-3,3,-3,3,-10,10);
		// */
		glMatrixMode(GL_MODELVIEW);
	}

	return (TRUE);

}

gint glarea_init(GtkWidget * widget)
{

	/* g_print("Realize Event\n"); */

	if (gtk_gl_area_make_current(GTK_GL_AREA(widget))) {

#define	GRAY(x)	x,x,x
		
		int i;

		glClearColor(GRAY(0.839f), 0.0f);
		glClearDepth(1.0);
		glDepthFunc(GL_LESS);
		glEnable(GL_DEPTH_TEST);
		glShadeModel(GL_SMOOTH);

		glGenTextures(6,&tex[0]);
		for(i=0;i<6;i++) {
			glBindTexture(GL_TEXTURE_2D,tex[i]);
			mktex(couleurs[i]);
			glTexImage2D(GL_TEXTURE_2D, 0, 3, TEX_SIZE, TEX_SIZE, 0, GL_RGB, GL_UNSIGNED_BYTE, texture);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
			free(texture);
		}
		glEnable(GL_TEXTURE_2D);
	}

	return TRUE;

}

gint glarea_destroy(GtkWidget * widget)
{

	// g_print("GTK GL Area Destroy Event\n");

	/* Insert any required cleanup */
	/* code here.                  */

	return TRUE;

}

gint animate(GtkWidget *glarea)
{
	int anim=0;
	if (!moving && pending_moves.size()) {
		// npending_moves--;
		// current_move=next_move;
		// next_move=(next_move+1) % MAX_MOVE_QUEUE;
		// moving=pending_moves[current_move].id;
		moving=pending_moves.front().id;
		ASSERT(moving & MOVING_MASK);
		current_move_rot=0;
	}
	if (moving) {
		if (pending_moves.front().nstep++ < MOVE_STEPS) {
			current_move_rot+=90.0 / MOVE_STEPS;
			anim=1;
		} else {
			apply_perm(pending_moves.front().id);
			/* g_print("Move completed !\n"); */
			moving=0;
			pending_moves.pop_front();
		}
	}
	if (anim)
		gtk_widget_draw(GTK_WIDGET(glarea), NULL);
	return TRUE;
}

guint16 clamp_16(double x)
{
	if (x > 65535.0)
		return 65535;
	else if (x<0)
		return 0;
	else
		return (guint16) x;
}


void set_bgcolor(GtkWidget * w, GtkStateType state, double mul, rgbcol col)
{
	GtkStyle *style;
	style = gtk_style_copy(gtk_widget_get_style(w));
	/*
	g_print("Previous bg on state %ld : %d %d %d %d\n",
			state,
			style->bg[state].pixel,
			style->bg[state].red,
			style->bg[state].green,
			style->bg[state].blue);
			*/
	style->bg[state].pixel	= 0;
	style->bg[state].red	= clamp_16(col[0] * mul * 65535.0);
	style->bg[state].green	= clamp_16(col[1] * mul * 65535.0);
	style->bg[state].blue	= clamp_16(col[2] * mul * 65535.0);
	gtk_widget_set_style(w, style);
}

int start_interface(int argc, char **argv)
{
	int i;

	GtkWidget *window;
	GtkWidget *utils;
	GtkWidget *box_main;
	GtkWidget *glarea;
	GtkWidget *button_box;
	GtkWidget *button_box2;


	gmp_randinit_default(random_source);
	gmp_randseed_ui(random_source,time(NULL));

	gtk_init(&argc, &argv);

	/* Main widget container */

	/* GTK GL Area */

	glarea = create_glarea();

	/***************************************************************/
	utils = gtk_hbutton_box_new();
	gtk_button_box_set_layout(GTK_BUTTON_BOX(utils),
				  GTK_BUTTONBOX_SPREAD);
	// gtk_button_box_set_child_size(GTK_BUTTON_BOX(utils), 30, 0);
	gtk_button_box_set_spacing(GTK_BUTTON_BOX(utils), 0);

	GtkWidget * button_random = gtk_button_new_with_label("Random");
	gtk_signal_connect(GTK_OBJECT(button_random), "clicked",
			   GTK_SIGNAL_FUNC(randomize), GTK_WIDGET(glarea));
	gtk_container_add(GTK_CONTAINER(utils),button_random);
	gtk_widget_show(button_random);

	GtkWidget * button_solve = gtk_button_new_with_label("Compute");
	gtk_signal_connect(GTK_OBJECT(button_solve), "clicked",
			   GTK_SIGNAL_FUNC(solve_cube), NULL);
	gtk_container_add(GTK_CONTAINER(utils),button_solve);
	gtk_widget_show(button_solve);

	GtkWidget * button_move = gtk_button_new_with_label("Advance");
	gtk_signal_connect(GTK_OBJECT(button_move), "clicked",
			   GTK_SIGNAL_FUNC(advance_on_solution), NULL);
	gtk_container_add(GTK_CONTAINER(utils),button_move);
	gtk_widget_show(button_move);

	GtkWidget * button_all = gtk_button_new_with_label("Do all");
	gtk_signal_connect(GTK_OBJECT(button_all), "clicked",
			   GTK_SIGNAL_FUNC(do_all_moves), NULL);
	gtk_container_add(GTK_CONTAINER(utils),button_all);
	gtk_widget_show(button_all);

	GtkWidget * button_quit = gtk_button_new_with_label("Quit");
	gtk_signal_connect(GTK_OBJECT(button_quit), "clicked",
			   GTK_SIGNAL_FUNC(gtk_main_quit), NULL);
	gtk_container_add(GTK_CONTAINER(utils),button_quit);
	gtk_widget_show(button_quit);

	/***************************************************************/
	button_box = gtk_hbutton_box_new();
	button_box2 = gtk_hbutton_box_new();

	gtk_button_box_set_layout(GTK_BUTTON_BOX(button_box),
				  GTK_BUTTONBOX_SPREAD);
	gtk_button_box_set_child_size(GTK_BUTTON_BOX(button_box), 40, 0);
	gtk_button_box_set_spacing(GTK_BUTTON_BOX(button_box), 0);

	gtk_button_box_set_layout(GTK_BUTTON_BOX(button_box2),
				  GTK_BUTTONBOX_SPREAD);
	gtk_button_box_set_child_size(GTK_BUTTON_BOX(button_box2), 40, 0);
	gtk_button_box_set_spacing(GTK_BUTTON_BOX(button_box2), 0);

	for (i = 0; i < 6; i++) {
		GtkWidget *button_i;
		int j;
		
		j=button_order[i];
		button_i = gtk_button_new_with_label(face_names[j]);
		gtk_signal_connect(GTK_OBJECT(button_i), "clicked",
				   GTK_SIGNAL_FUNC(handle_button),
				   &face_names[j]);
		set_bgcolor(button_i, GTK_STATE_NORMAL, 1.0, couleurs[j]);
		set_bgcolor(button_i, GTK_STATE_PRELIGHT, 1.0, couleurs[j]);
		set_bgcolor(button_i, GTK_STATE_ACTIVE, 0.9, couleurs[j]);
		gtk_container_add(GTK_CONTAINER(button_box), button_i);
		gtk_widget_show(button_i);

		button_i = gtk_button_new_with_label(face_names[j+6]);
		gtk_signal_connect(GTK_OBJECT(button_i), "clicked",
				   GTK_SIGNAL_FUNC(handle_button),
				   &face_names[j+6]);
		set_bgcolor(button_i, GTK_STATE_NORMAL, 1.0, couleurs[j]);
		set_bgcolor(button_i, GTK_STATE_PRELIGHT, 1.0, couleurs[j]);
		set_bgcolor(button_i, GTK_STATE_ACTIVE, 0.9, couleurs[j]);
		gtk_container_add(GTK_CONTAINER(button_box2), button_i);
		gtk_widget_show(button_i);
	}
	/* Main window */

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(window), "GtkGLArea Demo");

	/* destroy this window when exiting from gtk_main() */

	gtk_quit_add_destroy(1, GTK_OBJECT(window));

	gtk_signal_connect(GTK_OBJECT(window), "key_press_event",
			   GTK_SIGNAL_FUNC(key_press_event), NULL);
	gtk_signal_connect(GTK_OBJECT(window), "key_release_event",
			   GTK_SIGNAL_FUNC(key_release_event), NULL);

	gtk_signal_connect(GTK_OBJECT(window), "delete_event",
			   GTK_SIGNAL_FUNC(gtk_main_quit), NULL);

	gtk_signal_connect(GTK_OBJECT(window), "destroy",
			   GTK_SIGNAL_FUNC(gtk_main_quit), NULL);

	gtk_container_set_border_width(GTK_CONTAINER(window), 10);

	/* Pack everything into the main box, add the main box to */
	/* the top-level window, and 'show' all of the widgets.   */
	/* The window is 'shown' last so that when the user sees  */
	/* it, it's contents are already there. Otherwise they    */
	/* might see each widget come up.                         */

	box_main = gtk_vbox_new(FALSE, 10);
	gtk_box_pack_start(GTK_BOX(box_main), glarea, TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(box_main), utils, FALSE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(box_main), button_box, FALSE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(box_main), button_box2, FALSE, TRUE, 0);
	gtk_container_add(GTK_CONTAINER(window), box_main);

	gtk_widget_show(glarea);
	gtk_widget_show(button_box);
	gtk_widget_show(button_box2);
	gtk_widget_show(utils);
	gtk_widget_show(box_main);
	gtk_widget_show(window);

	GTK_WIDGET_SET_FLAGS(window, GTK_CAN_FOCUS);
	gtk_widget_grab_focus(GTK_WIDGET(window));

	gtk_timeout_add(10,(GtkFunction)animate, glarea);

	gtk_main();

	gmp_randclear(random_source);

	return (0);

}
